      MODULE swim_arrays
        REAL, DIMENSION(:,:), ALLOCATABLE :: U
        REAL, DIMENSION(:,:), ALLOCATABLE :: V
        REAL, DIMENSION(:,:), ALLOCATABLE :: P
        REAL, DIMENSION(:,:), ALLOCATABLE :: UNEW
        REAL, DIMENSION(:,:), ALLOCATABLE :: VNEW
        REAL, DIMENSION(:,:), ALLOCATABLE :: PNEW
        REAL, DIMENSION(:,:), ALLOCATABLE :: UOLD
        REAL, DIMENSION(:,:), ALLOCATABLE :: VOLD
        REAL, DIMENSION(:,:), ALLOCATABLE :: POLD
        REAL, DIMENSION(:,:), ALLOCATABLE :: CU
        REAL, DIMENSION(:,:), ALLOCATABLE :: CV
        REAL, DIMENSION(:,:), ALLOCATABLE :: Z
        REAL, DIMENSION(:,:), ALLOCATABLE :: H
        REAL, DIMENSION(:,:), ALLOCATABLE :: PSI
      END MODULE

      SUBROUTINE ALLOC_ARRAYS
        USE swim_arrays
        INCLUDE "swim.h"
        INTEGER status

        ALLOCATE(U(N1,N2), STAT=status)
        ALLOCATE(V(N1,N2), STAT=status)
        ALLOCATE(P(N1,N2), STAT=status)
        ALLOCATE(UNEW(N1,N2), STAT=status)
        ALLOCATE(VNEW(N1,N2), STAT=status)
        ALLOCATE(PNEW(N1,N2), STAT=status)
        ALLOCATE(UOLD(N1,N2), STAT=status)
        ALLOCATE(VOLD(N1,N2), STAT=status)
        ALLOCATE(POLD(N1,N2), STAT=status)
        ALLOCATE(CU(N1,N2), STAT=status)
        ALLOCATE(CV(N1,N2), STAT=status)
        ALLOCATE(Z(N1,N2), STAT=status)
        ALLOCATE(H(N1,N2), STAT=status)
        ALLOCATE(PSI(N1,N2), STAT=status)
      END SUBROUTINE

      PROGRAM SHALOW 
C     BENCHMARK WEATHER PREDICTION PROGRAM FOR COMPARING THE
C     PREFORMANCE OF CURRENT SUPERCOMPUTERS. THE MODEL IS
C     BASED OF THE PAPER - THE DYNAMICS OF FINITE-DIFFERENCE
C     MODELS OF THE SHALLOW-WATER EQUATIONS, BY ROBERT SADOURNY
C     J. ATM. SCIENCES, VOL 32, NO 4, APRIL 1975.
C
C     CODE BY PAUL N. SWARZTRAUBER, NATIONAL CENTER FOR
C     ATMOSPHERIC RESEARCH, BOULDER, CO,  OCTOBER 1984.
C
C     MODIFIED BY R. K. SATO, NCAR, APRIL 7, 1986 FOR MULTITASKING.
C     MODIFIED FOR SPEC to run longer: ITMAX inCremented from 120 to
C                                      1200 - J.Lo 7/19/90
C     Modified by Bodo Parady for the SPECpar suite.  Various
C     compilation sizes added.  Iterations reduced to orginal
C     problem, but size increased 4x in each dimension.
C
C     Further modified by Reinhold Weicker (Siemens Nixdorf) for the
C     SPEC CFP95 suite:
C     Changed back to the form with PARAMETER statements for
C     N1 and N2, set N1 = N2 = 513.
C     The execution time is determined by these values and the
C     variable ITMAX (number of iterations) read from the input file.
C     Execution time is linear in ITMAX.

      USE swim_arrays
C
      implicit none
C
      include "mpif.h"
      include "swim.h"
C
      integer i,j,mnmin,ncycle
      integer js,je,taskid,numtasks,req,istat,ierr
      integer M,N,MP1,NP1
      real TDT,EL,PI,TPI,DI,DJ,PCF
      integer JCHECK,ICHECK
      real TIME,PTIME,PCHECK,UCHECK,VCHECK
      real TPCHECK,TUCHECK,TVCHECK
C
      REAL*8 rtc,t1,t2
C
      COMMON/decomp/js,je,taskid,numtasks,req(16),
     1                            istat(MPI_STATUS_SIZE,16)
      COMMON /CONS/ TDT,M,N,MP1,
     1              NP1,EL,PI,TPI,DI,DJ,PCF
C
c      Get number of tasks working on this.

      cALL MPI_INIT( ierr )
 
      cALL MPI_COMM_RANK( MPI_COMM_WORLD, taskid, ierr )
      cALL MPI_COMM_SIZE( MPI_COMM_WORLD, numtasks, ierr )
      if(taskid.eq.MASTER)then
c       timei =rtc()
        WRITE(6,*) 'SPEC benchmark 102.swim'
        WRITE(6,*) ' '
      endif
  
C
C Initializing array allocation
C
      call alloc_arrays
      
C
C Initializing array req with MPI_REQUEST_NULL
C
      do i=1,16
	req(i) = MPI_REQUEST_NULL
      end do

C       REQUEST PROCESSORS FOR MICROTASKING
C SPEC removed CMIC$ GETCPUS 4
C
C
C       INITIALIZE CONSTANTS AND ARRAYS
C
      CALL INITAL
C
C     PRINT INITIAL VALUES
C
      if(taskid.eq.MASTER)then
      WRITE(6,390) N,M,DX,DY,DT,ALPHA,ITMAX
  390 FORMAT(' NUMBER OF POINTS IN THE X DIRECTION', I8/
     1       ' NUMBER OF POINTS IN THE Y DIRECTION', I8/
     2       ' GRID SPACING IN THE X DIRECTION    ', F8.0/
     3       ' GRID SPACING IN THE Y DIRECTION    ', F8.0/
     4       ' TIME STEP                          ', F8.0/
     5       ' TIME FILTER PARAMETER              ', F8.3/
     6       ' NUMBER OF ITERATIONS               ', I8)
      endif
      MNMIN = MIN0(M,N)
C initial data writes removed for SPEC
C      WRITE(6,391) (POLD(I,I),I=1,MNMIN)
C  391 FORMAT(/' INITIAL DIAGONAL ELEMENTS OF P ', //(8E15.7))
C      WRITE(6,392) (UOLD(I,I),I=1,MNMIN)
C  392 FORMAT(/' INITIAL DIAGONAL ELEMENTS OF U ', //(8E15.7))
C      WRITE(6,393) (VOLD(I,I),I=1,MNMIN)
C  393 FORMAT(/' INITIAL DIAGONAL ELEMENTS OF V ', //(8E15.7))
C        DETERMINE OVERHEAD OF TIMING CALLS

C  6/22/95 for SPEC: JWR: Initialization of TIME
      TIME = 0
      NCYCLE = 0
c     t1 = rtc()
C
Ccc beginning of do 90 loop
C
      DO NCYCLE = 1,ITMAX
C
C     COMPUTE CAPITAL  U, CAPITAL V, Z AND H
C
      CALL CALC1
C
C     COMPUTE NEW VALUES U,V AND P
C
      CALL CALC2
C
      TIME = TIME + DT
      IF(MOD(NCYCLE,MPRINT) .eq. 0) then
      
      PTIME = TIME/3600.
C
C *** modified for SPEC results verification
C *** We want to ensure that ALL calculations were done
C *** so we "use" ALL of the computed results.
C ***
C *** Since ALL of the comparisons of the individual diagnal terms
C *** often differ in the smALLer values, the check we have selected
C *** is to add the absolute values of ALL terms and print these results
C
  350 FORMAT(/' CYCLE NUMBER',I5,' MODEL TIME IN  HOURS', F6.2)
C      WRITE(6,355) (PNEW(I,I),I=1,MNMIN)
C  355 FORMAT(/' DIAGONAL ELEMENTS OF P ', //(8E15.7))
C      WRITE(6,360) (UNEW(I,I),I=1,MNMIN)
C  360 FORMAT(/' DIAGONAL ELEMENTS OF U ', //(8E15.7))
C      WRITE(6,365) (VNEW(I,I),I=1,MNMIN)
C  365 FORMAT(/' DIAGONAL ELEMENTS OF V ', //(8E15.7))
C ***
        PCHECK = 0.0
        UCHECK = 0.0
        VCHECK = 0.0
        
        DO 3500 JCHECK = js,je
         DO 3500 ICHECK = 1, MNMIN
         PCHECK = PCHECK + ABS(PNEW(ICHECK,JCHECK))
         UCHECK = UCHECK + ABS(UNEW(ICHECK,JCHECK))
         VCHECK = VCHECK + ABS(VNEW(ICHECK,JCHECK))
 3500   CONTINUE
        cALL mpi_reduce(pcheck,tpcheck,1,MPI_DOUBLE_PRECISION,
     1    MPI_SUM,0,MPI_COMM_WORLD,ierr)
        cALL mpi_reduce(ucheck,tucheck,1,MPI_DOUBLE_PRECISION,
     1    MPI_SUM,0,MPI_COMM_WORLD,ierr)
        cALL mpi_reduce(vcheck,tvcheck,1,MPI_DOUBLE_PRECISION,
     1    MPI_SUM,0,MPI_COMM_WORLD,ierr)
C ***
C ***
      if(taskid.eq.MASTER)then
        WRITE(6,350) NCYCLE,PTIME
        WRITE(6,366) tPCHECK, tUCHECK, tVCHECK
      endif
 366    FORMAT(/,
     *         ' Pcheck = ',E12.4,/,
     *         ' Ucheck = ',E12.4,/,
     *         ' Vcheck = ',E12.4,/)

      END IF
C
C     TIME SMOOTHING AND UPDATE FOR NEXT CYCLE
C
      IF(NCYCLE .LE. 1) THEN
        CALL CALC3Z
      ELSE IF(NCYCLE .lt. ITMAX) THEN
        CALL CALC3
      ENDIF
C
C end of goto 90 loop
C
      END DO
C
      CALL MPI_FINALIZE(ierr)
      STOP
      END
C
C
C
C
C
      SUBROUTINE INITAL
C        INITIALIZE CONSTANTS AND ARRAYS
C           R. K. SATO 4/7/86
C
      USE swim_arrays
C
      implicit none
C
      include "mpif.h"
      include "swim.h"
C
      integer i,j,k
      integer M,N,MP1,NP1
      real TDT,EL,PI,TPI,DI,DJ,PCF
      integer js,je,taskid,numtasks,req,istat,ierr
      integer left,right,endtask
C
      COMMON/decomp/js,je,taskid,numtasks,req(MAX_REQ),
     1                      istat(MPI_STATUS_SIZE,MAX_REQ)
C
      COMMON /CONS/ TDT,M,N,MP1,
     1              NP1,EL,PI,TPI,DI,DJ,PCF
C
C     Find neighbors
C
      endtask = numtasks-1
      if (taskid.eq.MASTER)then
        left = MPI_PROC_NULL
      else
        left = taskid-1
      endif
      if (taskid.eq.endtask)then
        right = MPI_PROC_NULL
      else
        right = taskid+1
      endif
C
C     NOTE BELOW THAT TWO DELTA T (TDT) IS SET TO DT ON THE FIRST
C     CYCLE AFTER WHICH IT IS RESET TO DT+DT.
C
C The following code  was in SWM256, however, it was replaced by 
C		READ statements to avoid calculations to be done during
C		compile time.
C
       M = N1 - 1
       N = N2 - 1
c     READ (5,*) DT
c     READ (5,*) DX
c     READ (5,*) DY
c     READ (5,*) A
c     READ (5,*) ALPHA
c     READ (5,*) ITMAX
c     READ (5,*) MPRINT
c     READ (5,*) M
c     READ (5,*) N
      TDT = DT
      MP1 = M+1
      NP1 = N+1
      EL = REAL(N)*DX
      PI = 4.*ATAN(1.)
      TPI = PI+PI
      DI = TPI/REAL(M)
      DJ = TPI/REAL(N)
      PCF = PI*PI*A*A/(EL*EL)
      js = 1 + (N/(numtasks+1))*taskid
      je = min(n2,js+N/numtasks)
C
C     INITIAL VALUES OF THE STREAM FUNCTION AND P
C
      DO J=js,je
        DO I=1,MP1
          PSI(I,J) = A*SIN((REAL(I)-.5)*DI)*SIN((REAL(J)-.5)*DJ)
          P(I,J) = PCF*(COS(2.*REAL(I-1)*DI)
     1                +COS(2.*REAL(J-1)*DJ))+50000.
          if(j.eq.n)then
            P(I,J+1) = PCF*(COS(2.*REAL(I-1)*DI)
     1                +COS(2.*REAL(J)*DJ))+50000.
          endif
        END DO
      END DO
C
C
C     INITIALIZE VELOCITIES
C
      if(taskid.lt.endtask)then
        cALL mpi_irecv(psi(1,je+1),n1,MPI_DOUBLE_PRECISION,
     1    right,1,MPI_COMM_WORLD,req(1),ierr)
      endif
      if(taskid.gt.MASTER)then
        cALL mpi_isend(psi(1,js),n1,MPI_DOUBLE_PRECISION,
     1    left,1,MPI_COMM_WORLD,req(2),ierr)
      endif
      cALL MPI_WAITALL(2,req,istat,ierr)
      DO J=js,je
        DO I=1,M
          U(I+1,J) = -(PSI(I+1,J+1)-PSI(I+1,J))/DY
          if(j.gt.1)then
            V(I,J) = (PSI(I+1,J)-PSI(I,J))/DX
          endif
          if(j.eq.n)then
            V(I,J+1) = (PSI(I+1,J+1)-PSI(I,J+1))/DX
          endif
        END DO
      END DO
C
C     PERIODIC CONTINUATION
C
      DO J=1,N
        U(1,J) = U(M+1,J)
        V(M+1,J) = V(1,J)
        if(j.eq.n)then
          V(M+1,J+1) = V(1,J+1)
        endif
      END DO
      if(taskid.eq.MASTER)then
        cALL mpi_irecv(v(1,1),n1,MPI_DOUBLE_PRECISION,
     1    endtask,1,MPI_COMM_WORLD,req(1),ierr)
      endif
      if(taskid.eq.endtask)then
        cALL mpi_irecv(u(1,n+1),n1,MPI_DOUBLE_PRECISION,
     1    MASTER,2,MPI_COMM_WORLD,req(2),ierr)
      endif
      if(taskid.eq.endtask)then
        cALL mpi_isend(v(1,n+1),n1,MPI_DOUBLE_PRECISION,
     1    MASTER,1,MPI_COMM_WORLD,req(3),ierr)
      endif
      if(taskid.eq.MASTER)then
        cALL mpi_isend(u(1,1),n1,MPI_DOUBLE_PRECISION,
     1    endtask,2,MPI_COMM_WORLD,req(4),ierr)
      endif
      if(taskid.eq.MASTER.or.taskid.eq.endtask)then
        cALL MPI_WAITALL(4,req,istat,ierr)
      endif

c     DO 75 I=1,M
c     U(I+1,N+1) = U(I+1,1)
c     V(I,1) = V(I,N+1)
c  75 CONTINUE
cm tag_v=2
      if(taskid.eq.MASTER)then
        cALL mpi_irecv(v(m+1,1),1,MPI_DOUBLE_PRECISION,
     1    endtask,1,MPI_COMM_WORLD,req(1),ierr)
      endif
      if(taskid.eq.endtask)then
        cALL mpi_irecv(u(1,n+1),1,MPI_DOUBLE_PRECISION,
     1    MASTER,2,MPI_COMM_WORLD,req(2),ierr)
      endif
      if(taskid.eq.endtask)then
        cALL mpi_isend(v(1,n+1),1,MPI_DOUBLE_PRECISION,
     1    MASTER,1,MPI_COMM_WORLD,req(3),ierr)
      endif
      if(taskid.eq.MASTER)then
        cALL mpi_isend(u(M+1,1),1,MPI_DOUBLE_PRECISION,
     1    endtask,2,MPI_COMM_WORLD,req(4),ierr)
      endif
      if(taskid.eq.MASTER.or.taskid.eq.endtask)then
        cALL MPI_WAITALL(4,req,istat,ierr)
      endif

c     U(1,N+1) = U(M+1,1)
c     V(M+1,1) = V(1,N+1)
      DO J=js,je
        DO I=1,MP1
          UOLD(I,J) = U(I,J)
          VOLD(I,J) = V(I,J)
          POLD(I,J) = P(I,J)
          if(j.eq.n)then
            UOLD(I,J+1) = U(I,J+1)
            VOLD(I,J+1) = V(I,J+1)
            POLD(I,J+1) = P(I,J+1)
          endif
        END DO
      END DO
C        END OF INITIALIZATION
      RETURN
      END SUBROUTINE
c
c
c
